@@ -1,10 +1,11 @@
 #!/usr/bin/env perl
 
 # Flex(1) XML processor scanner generator.
-# Copyright  1999 Kristoffer Rose.  All rights reserved.
+# Copyright (C) 1999 Kristoffer Rose.  All rights reserved.
 #
 # This file is part of the FleXML XML processor generator system.
-# Copyright  1999 Kristoffer Rose.  All rights reserved.
+# Copyright (C) 1999 Kristoffer Rose.  All rights reserved.
+# Copyright (C) 2003-2006 Martin Quinson.  All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the Free
@@ -22,14 +23,14 @@
 
 use warnings;
 
-my $Id = '$Id: flexml.pl,v 1.44 2005/02/23 23:08:16 mquinson Exp $ ';
+my $Id = '$Id: flexml.pl,v 1.61 2006/09/13 16:34:33 wdowling Exp $ ';
 $Id =~ s/\s*\$\s*//go;
 
 # IMPORTS.
 
-use Getopt::Std;
+use Getopt::Long;
 use LWP::UserAgent;
-use Carp qw/cluck/;
+use Carp qw/cluck confess/;
 
 use strict;
 
@@ -41,10 +42,11 @@ use vars qw/ $HEADER /;		# generated XML processor API header
 use vars qw/ $ACTIONS /;	# the actions file
 use vars qw/ $APPLICATION /;	# generated XML application
 
+$SKELETON = "./skel"; # default value
+
 # OPTIONS (and other globals).
 
 my $Use;			# usage string
-my %opt = ();			# options
 
 my $debug;			# -d option flag
 my $verbose;			# -v option flag
@@ -53,7 +55,19 @@ my $nofail;			# -X option flag
 my $quiet_parser;		# -q option flag
 my $uri;			# -u option uri
 my $pubid;			# -p option string
-my $stacksize;			# -b option flag
+my $sysid;			# --sysid option string
+my $stacksize=100000;		# -b option flag
+my $tagprefix="";      		# -P option flag
+my $actbin="./flexml-act";      # -T option content
+my $init_header="";             # -i option string
+
+my $header;			# -H option flag/content
+my $dummy;			# -D option flag/content
+my $standalone;			# -A option flag/content
+my $scanner;			# -S option flag/content
+my $actions;			# -a option content
+
+my $dryrun;                     # -n option flag
 
 my $dtd;			# DTD file name (or URI)
 my $dtdrevision;		# DTD version pruned from file
@@ -164,7 +178,8 @@ sub redistribute { # Print C comment with generated file "license".
   my ($pre) = @_;
   print <<EOT;
 $pre This program was generated with the FleXML XML processor generator.
-$pre FleXML is Copyright  1999-2005 Kristoffer Rose.  All rights reserved.
+$pre FleXML is Copyright (C) 1999-2005 Kristoffer Rose.  All rights reserved.
+$pre FleXML is Copyright (C) 2003-2006 Martin Quinson.  All rights reserved.
 $pre ($Id).
 $pre 
 $pre There are two, intertwined parts to this program, part A and part B.
@@ -173,8 +188,8 @@ $pre Part A
 $pre ------
 $pre 
 $pre Some parts, here collectively called "Part A", are found in the 
-$pre FleXML package.  They are Copyright  1999-2005 Kristoffer Rose. 
-$pre All rights reserved.
+$pre FleXML package.  They are Copyright (C) 1999-2005 Kristoffer Rose
+$pre and Copyright (C) 2003-2006 Martin Quinson. All rights reserved.
 $pre
 $pre You can redistribute, use, perform, display and/or modify "Part A"
 $pre provided the following two conditions hold:
@@ -212,8 +227,8 @@ sub api_functions { # Print XML application interface functions.
   my ($pre,$post) = @_;
   print "/* XML application entry points. */\n" if @tags;
   for (@tags) {
-    print $pre . "void STag_$ctag{$_}(void)$post\n" unless $startok{$_};
-    print $pre . "void ETag_$ctag{$_}(void)$post\n" unless $endok{$_};
+    print $pre . "void STag_${tagprefix}$ctag{$_}(void)$post\n" unless $startok{$_};
+    print $pre . "void ETag_${tagprefix}$ctag{$_}(void)$post\n" unless $endok{$_};
   }
 }
 
@@ -223,8 +238,8 @@ sub api_types { # Print XML application interface types.
   for (keys %atttype) {
     if (m.($Nmtoken)[/]($Nmtoken).xo) {
       my ($tag,$attribute) = ($1,$2);
-      print "typedef $typeof{$_} AT_$ctag{$tag}_$catt{$attribute};\n";
-      print "#define AU_$ctag{$tag}_$catt{$attribute} NULL\n"
+      print "typedef $typeof{$_} AT_${tagprefix}$ctag{$tag}_$catt{$attribute};\n";
+      print "#define AU_${tagprefix}$ctag{$tag}_$catt{$attribute} NULL\n"
 	if not $enumtype{$_};
     }
   }
@@ -232,12 +247,22 @@ sub api_types { # Print XML application interface types.
 
 
 sub api_data { # Print XML application interface parameters.
-  my ($pre) = @_;
+  my ($pre) = @_; # pre should be a storage class spec like 'static' or 'extern'
   print "/* FleXML-provided data. */\n";
-  print $pre . "const char* pcdata;\n";
+  print $pre . "int ${tagprefix}pcdata_ix;\n";
+  print "extern char *${tagprefix}bufferstack;\n";
+  print "#define ${tagprefix}pcdata (${tagprefix}bufferstack + ${tagprefix}pcdata_ix)\n";
+
   for (keys %atttype) {
     if (m.($Nmtoken)[/]($Nmtoken).xo) {
-      print $pre . "AT_$ctag{$1}_$catt{$2} A_$ctag{$1}_$catt{$2};\n";
+      print $pre . "AT_${tagprefix}$ctag{$1}_$catt{$2} AX_${tagprefix}$ctag{$1}_$catt{$2};\n";
+      print "#define A_${tagprefix}$ctag{$1}_$catt{$2} ";
+      if ($enumtype{$_}) {
+	  print "AX_${tagprefix}$ctag{$1}_$catt{$2}\n";
+      }
+      else {
+	  print "(${tagprefix}bufferstack + AX_${tagprefix}$ctag{$1}_$catt{$2})\n";
+      }
     }
   }
 }
@@ -246,7 +271,7 @@ sub api_data { # Print XML application interface parameters.
 sub expandparametersat { # Expand parameter entities in $_ at $place.
   my ($place) = @_;
   while ( m/$place\s*%($Name);/ ) {
-    if ($parameter{$1}) {
+    if (defined $parameter{$1}) {
       s/$place(\s*)%($Name);/ ' '. $1 . $parameter{$2} . ' ' /xe;
     }
   }
@@ -298,6 +323,16 @@ sub geturl { # Insert contents of URL into input stream at current point
 #  }
 }
 
+# add attribute to list of all default attributes
+my @default_attributes;
+my $next_att_loc = 1;
+sub add_def_att {
+    my ($att) = @_;
+    my $retval = $next_att_loc;
+    push @default_attributes, ($att);
+    $next_att_loc += length($att) + 1;
+    return $retval;
+}
 
 sub nextline {			# return one input line
 
@@ -485,43 +520,71 @@ sub analysechildren { # Analyse DTD children specification; return
 
 
 # Parse options.
-$Use = "Usage: flexml [-ASHDvdqnLXV] [-s skel] [-p pubid] [-u uri]\n"
-     . "       [-b stack_size] [-r roottags] [-a actions] name[.dtd]";
+$Use = "Usage: flexml [-ASHDvdqnLXV] [-s skel] [-T actbin] [--sysid sysid] "
+  . "[-p pubid] [-u uri] [-i init_header]\n"
+  . "       [-b stack_increment] [-r roottags] [-a actions] [-P prefix] name[.dtd]";
 
-getopts('ASHDvdnLXVqp:b:s:u:r:a:', \%opt);
+sub show_version {
 
-# Version!
-print "FleXML version $Id.\n" if $opt{V} or $opt{v};
-exit 0 if $opt{V};
-
-# Debugging?
-$debug = $opt{d};
-$verbose = $opt{v};
-
-# Line numbers?
-$lineno = $opt{L};
-
-# Quiet parser?    
-$quiet_parser = $opt{q};
-    
-# Exit without fail message?
-$nofail = $opt{X};
-
-# Specific root tags?
-if ($opt{r}) {
-  for (split ',',$opt{r}) { $roottags{$_} = 'true'; }
+    exit 0;
 }
+    
+Getopt::Long::Configure ("bundling");
+GetOptions(
+    # Debugging and verbosity
+    "debug|d" => \$debug,
+    "verbose|v" => \$verbose,
+    "quiet|q" => \$quiet_parser,
+    
+    # Version!
+    "version|V" => sub { print "FleXML version $Id.\n"; exit 0; },
 
-# Specific stack size?
-$stacksize = ($opt{'b'} ? $opt{'b'} : 100000);
+    # dry-run ?
+    "dry-run|n" => \$dryrun,
 
-# Set skeleton scanner file name and check it is readable (if needed).
-$SKELETON = ($opt{'s'} ? $opt{'s'} : './skel');
-die "$0: No skeleton file $SKELETON.\n" if not -r $SKELETON and $opt{S};
+    # Line numbers?
+    "lineno|L" => \$lineno,
+    
+    # Exit without fail message?
+    "nofail|X" => \$nofail,
 
-# Set document type URI and PUBID.
-$uri = $opt{u} if $opt{u};
-$pubid = $opt{p} if $opt{p};
+    # Specific root tags?
+    "root-tags|r=s" => sub {
+	for (split ',',$_[1]) { $roottags{$_} = 'true'; }
+    },
+
+    # Specific stack size?
+    "stack-increment|b=s" => \$stacksize,
+    
+    # Specific tagprefix?
+    "tag-prefix|P=s" => sub { $tagprefix = $_[1]."_" },
+    
+    # Specific actbin? (internal use)
+    "act-bin|T=s" => \$actbin,
+    
+    # Set skeleton scanner file name and check it is readable (if needed).
+    "skel|s=s" => sub {
+	$SKELETON = $_[1];
+	die "$0: No skeleton file $SKELETON.\n" if not -r $SKELETON and $_[1];
+    },
+    
+    # Set document type URI and PUBID.
+    "uri|u=s" => \$uri,
+    "pubid|p=s" => \$pubid,
+    "sysid=s" => \$sysid,
+
+    # name of header file to be included in initial section of generated .l file
+    "init_header|i=s" => \$init_header,
+
+    # What to generate
+    "header|H:s" => sub { $header = $_[1] || 'true' },
+    "dummy|D:s" => sub { $dummy = $_[1] || 'true' },
+    "stand-alone|A" => \$standalone,
+    "scanner|S:s" => sub { $scanner = $_[1] || 'true' },
+    "actions|a=s" => \$actions
+    );
+
+print "FleXML version $Id.\n" if $verbose;
 
 # Set DTD file name...and extract prefix for later
 my $prefix = $ARGV[0];
@@ -538,30 +601,30 @@ else {
 
 # Selection options: If none of -SHDA specified then default to -SH.
 # Furthermore -a implies -D.
-$opt{S} = $opt{H} = 'true' unless ($opt{S} or $opt{H} or $opt{D} or $opt{A});
-$opt{D} ||= $opt{a} unless $opt{A};
+$scanner = $header = 'true' unless ($scanner or $header or $dummy or $standalone);
+$dummy ||= $actions unless $standalone;
 
 # Set default (DTD-based) output file names.
-$SCANNER = "$prefix.l";
-$HEADER =  "$prefix.h";
-$APPLICATION = "$prefix-dummy.c";
+$SCANNER =     (!defined($scanner)) || $scanner eq 'true' ? "$prefix.l"       : $scanner;
+$HEADER =      (!defined($header))  || $header eq 'true'  ? "$prefix.h"       : $header;
+$APPLICATION = (!defined($dummy))   || $dummy eq 'true'   ? "$prefix-dummy.c" : $dummy;
 
 # Set actions=based output file names, if any.
-if ($ACTIONS = $opt{a}) {
-  $opt{a} =~ s/\.[a-z]+$//;
-  $APPLICATION = "$opt{a}.c";
+if ($ACTIONS = $actions) {
+  $actions =~ s/\.[a-z]+$//;
+  $APPLICATION = "$actions.c";
 }
 
 # Stand-alone applications...
-if ($opt{A}) {
-  die "$0: -A conflicts with -SHD.\n" if ($opt{S} or $opt{H} or $opt{D});
+if ($standalone) {
+  die "$0: -A conflicts with -SHD.\n" if ($scanner or $header or $dummy);
   $SCANNER = $APPLICATION;
   $SCANNER =~ s/\.c$/.l/;
 }
 
 # Dry-run?
-if ($opt{n}) {
-  $opt{A} = $opt{S} = $opt{H} = $opt{D} = undef;
+if ($dryrun) {
+  $standalone = $scanner = $header = $dummy = undef;
 }
 
 
@@ -664,11 +727,12 @@ while (@inputs) {		# While there are lines...
     # strip off the matched code from the beginning of $_
     $_ = $'; #'
 
+    # place the newly found tag into the list of tags @tags
+    die place()."Repeated element $tag.\n" if exists $source{$tag};
+
     # ????? - most of the time $orig is undef
     $source{$tag} = "$orig";
 
-    # place the newly found tag into the list of tags @tags
-    die place()."Repeated element $tag.\n" if $ctag{$tag};
     push @tags, $tag;
 
     # Create C-friendly tag names.
@@ -682,7 +746,7 @@ while (@inputs) {		# While there are lines...
     # All elements should be followed by nothing when at the root.
     # IF there is a list of roottags ($0 -r ..,..), then only add
     # the exittrans if it is in the list
-    $exittrans{$tag} .= ($exittrans{$tag}?',':'') . "ROOT_$c=>EPILOG"
+    $exittrans{$tag} .= ($exittrans{$tag}?',':'') . "ROOT_${tagprefix}$c=>EPILOG"
       if not %roottags or $roottags{$tag};
 
     # Handle element declaration.
@@ -792,13 +856,13 @@ while (@inputs) {		# While there are lines...
 	    s/\s+//go;
 	    s/\|/,/go;
 	    $enumtype{"$tag/$attribute"} = "$_";
-	    s/$Nmtoken/ "A_$ctag{$tag}_$catt{$attribute}_" . variablify($&) /xge;
-	    my $undefined = "AU_$ctag{$tag}_$catt{$attribute}";
+	    s/$Nmtoken/ "A_${tagprefix}$ctag{$tag}_$catt{$attribute}_" . variablify($&) /xge;
+	    my $undefined = "AU_${tagprefix}$ctag{$tag}_$catt{$attribute}";
 	    s/^/enum \{ $undefined, /o;
 	    s/$/ \}/o;
 	    $typeof{"$tag/$attribute"} = "$_";
 	    if ($default) {
-	      $initof{"$tag/$attribute"} = "A_$ctag{$tag}_$catt{$attribute}_"
+	      $initof{"$tag/$attribute"} = "A_${tagprefix}$ctag{$tag}_$catt{$attribute}_"
 					 . variablify($default);
 	    }
 	    else {
@@ -807,12 +871,12 @@ while (@inputs) {		# While there are lines...
 	  }
 	  # ...and string/token types.
 	  else {
-	    $typeof{"$tag/$attribute"} = 'const char*';
+	    $typeof{"$tag/$attribute"} = 'int';
 	    if ($default) {
-	      $initof{"$tag/$attribute"} = "\"$default\"";
+	      $initof{"$tag/$attribute"} = add_def_att($default);
 	    }
 	    else {
-	      $initof{"$tag/$attribute"} = 'NULL';
+	      $initof{"$tag/$attribute"} = '0';
 	    }
 
 	    # Special treatment of token types.
@@ -831,10 +895,15 @@ while (@inputs) {		# While there are lines...
 	      $literaltype{"$tag/$attribute"} = '{Nmtokens}';
 	    }
 	    elsif ( $type eq 'ENTITY' ) {
-	      die place()."ENTITY attribute type unimplemented.\n";
+	      #die place()."ENTITY attribute type unimplemented.\n";
+	      $literaltype{"$tag/$attribute"} = '{Name}';
+	      print STDERR place()."Warning: attribute type `$type' not validated.\n";
 	    }
 	    elsif ( $type eq 'ENTITIES' ) {
-	      die place()."ENTITIES attribute type unimplemented.\n";
+	      #die place()."ENTITIES attribute type unimplemented.\n";
+	      $literaltype{"$tag/$attribute"} = '{Names}';
+	      print STDERR place()."Warning: attribute type `$type' not validated.\n";
+
 	    }
 	    elsif ( $type ne 'CDATA' ) {
 	      die place()."Unknown AttType `$type'.\n";
@@ -971,6 +1040,23 @@ if (not %roottags) {
   for (@tags) { $roottags{$_} = 'true'; }
 }
 
+# Handling prefix
+if(length($tagprefix)) {
+    my($h,$k);
+    my(@hashlist) = (\%states,\%emptytrans,\%instates,\%startstate,
+		     \%endstates,\%exittrans);
+
+    foreach $h (@hashlist) {
+	foreach $k (keys (%$h)) {
+	    $$h{$k} =~ s/^E_/E_${tagprefix}/g;
+	    $$h{$k} =~ s/,E_/,E_${tagprefix}/g;
+	    $$h{$k} =~ s/>E_/>E_${tagprefix}/g;
+	    $$h{$k} =~ s/^S_/S_${tagprefix}/g;
+	    $$h{$k} =~ s/,S_/,S_${tagprefix}/g;
+	    $$h{$k} =~ s/>S_/>S_${tagprefix}/g;
+	}
+    }
+}
 
 # Debugging: show DTD representation.
 
@@ -984,15 +1070,39 @@ if ($debug) {
     my ($k,$v);
     my $out = '';
 
-    while ( ($k, $v) = each(%$name) )
-    {
-      $out .= "\n    $k => " . ($v || "undef"); 
-  }
+    while ( ($k, $v) = each(%$name) ) {
+	$out .= "\n    $k => '" . ($v || "undef")."'"; 
+    }
 
     return $out;
   }
 
-  print '%opt = (' . printhash(\%opt) . ")\n";
+  # display the options we got
+  print "debug=".($debug||'undef')."\n";
+  print "verbose=".($verbose||'undef')."\n";
+  print "quiet=".($quiet_parser||'undef')."\n";
+
+  print "dry-run=".($dryrun||'undef')."\n";
+  print "lineno=".($lineno||'undef')."\n";
+
+  print "nofail=".($nofail||'undef')."\n";
+
+  print "stack-increment=".($stacksize||'undef')."\n";
+  print "tag-prefix=".($tagprefix||'undef')."\n";
+
+  print "act-bin=".($actbin||'undef')."\n";
+
+  print "skel=".($SKELETON||'undef')."\n";
+
+  print "uri=".($uri||'undef')."\n";
+  print "pubid=".($pubid||'undef')."\n\n";
+  print "sysid=".($sysid||'undef')."\n\n";
+
+  print "header=".($header||'undef')."\n";
+  print "dummy=".($dummy||'undef')."\n";
+  print "stand-alone=".($standalone||'undef')."\n";
+  print "scanner=".($scanner||'undef')."\n";
+  print "actions=".($actions||'undef')."\n\n";
 
   print '%source = (' . printhash(\%source) . ")\n";
   print "\n";
@@ -1041,8 +1151,7 @@ if ($debug) {
 
 # WRITE API HEADER (if requested).
 
-
-if ($opt{H}) {
+if ($header) {
 
   print STDOUT "Generating XML processor header in `$HEADER'.\n" if $verbose;
 
@@ -1070,14 +1179,14 @@ if ($opt{H}) {
   print "\n";
 
   print "/* XML application utilities. */\n";
-  print "extern int element_context(int);\n";
+  print "extern int ${tagprefix}element_context(int);\n";
   print "\n";
   print "/* XML processor entry point. */\n";
   print "extern int yylex(void);\n";
   print "\n";
   
   print "/* Flexml error handling function (useful only when -q flag passed to flexml) */\n";
-  print "const char * parse_err_msg(void);\n";
+  print "const char * ${tagprefix}parse_err_msg(void);\n";
     
   print "#endif\n";
 
@@ -1089,10 +1198,10 @@ if ($opt{H}) {
 # WRITE XML PROCESSOR (if requested).
 
 
-if ($opt{S} or $opt{A}) {
+if ($scanner or $standalone) {
 
   print STDOUT "Writing XML processor"
-             . ($opt{a} || $opt{A} ? " and application" : "")
+             . ($actions || $standalone ? " and application" : "")
              . " onto `$SCANNER'.\n" if $verbose;
 
   open SCANNER, "+>$SCANNER"|| die "$0: cannot write $SCANNER: $!\n";
@@ -1118,7 +1227,7 @@ if ($opt{S} or $opt{A}) {
 
     if ( /^FLEXML_VERSION$/ ) {
 
-      print "const char rcs_flexml[] =\n"
+      print "const char rcs_${tagprefix}flexml[] =\n"
 	  . " \"\$\" \"$Id \$\";\n";
 
       print "const char rcs_${cdtd}_dtd[] =\n"
@@ -1131,11 +1240,14 @@ if ($opt{S} or $opt{A}) {
       print "#define FLEXML_yylineno\n" if $lineno;
       print "#define FLEXML_NOFAIL\n" if $nofail;
       print "#define FLEXML_quiet_parser\n" if $quiet_parser;
-      print "#define FLEXML_HasMixed\n" if %inmixed;
+      print "#ifndef FLEXML_BUFFERSTACKSIZE\n";
       print "#define FLEXML_BUFFERSTACKSIZE $stacksize\n";
+      print "#endif\n";
+      print "#define FLEXML_NEED_BUFFERLIT\n"
+	if (scalar(%literaltype) or ($#default_attributes >= 0));
       print "\n";
 
-      if ($opt{A}) {
+      if ($standalone) {
 	api_functions('static ',';');
 	print "\n";
 	api_types();
@@ -1145,11 +1257,18 @@ if ($opt{S} or $opt{A}) {
       else {
 	print "/* XML processor api. */\n";
 	print "#include \"$HEADER\"\n\n"
-	    if ($opt{H});
+	    if ($header);
 	api_data('');
       }
 
     }
+    elsif ( /^FLEXML_INCLUDE_INIT_HEADER$/ ) {
+	if ($init_header) {
+	    print "/* User-supplied header */\n";
+	    print "#include \"$init_header\"\n";
+	}
+    }
+
     elsif ( /^FLEXML_FLEX_OPTIONS$/ ) {
 
       print "%option yylineno\n" if $lineno;
@@ -1162,8 +1281,8 @@ if ($opt{S} or $opt{A}) {
       for (@tags) {
 	my $c = $ctag{$_};
 	print "%x"
-            . ($roottags{$_} ? " ROOT_$c" : "")
-            . " AL_$c " . join(' ',split(',',$states{$_})) . "\n";
+            . ($roottags{$_} ? " ROOT_${tagprefix}$c" : "")
+            . " AL_${tagprefix}$c " . join(' ',split(',',$states{$_})) . "\n";
       }
 
     }
@@ -1171,7 +1290,7 @@ if ($opt{S} or $opt{A}) {
 
      print "%{\n";
      print "/* State names. */\n";
-     print "const char* *statenames=NULL;\n";
+     print "const char* *${tagprefix}statenames=NULL;\n";
      print "%}\n";
 
     }
@@ -1180,34 +1299,38 @@ if ($opt{S} or $opt{A}) {
       my ($state, $tag);
 
       print "  /* FleXML_init */\n";
-      print "  next = bufferstack;\n";
-      print "  if(!statenames) {statenames= (const char **)calloc(IMPOSSIBLE,sizeof(char*));\n";
+      print "  bnext = inext = 1;\n";
+      for my $att (@default_attributes) {
+	  print "  ${tagprefix}bufferliteral('\\0', &bnext, \"$att\");\n";
+      }
+      print "  if(!${tagprefix}statenames) {${tagprefix}statenames= (const char **)calloc(IMPOSSIBLE,sizeof(char*));\n";
       for ('PROLOG','DOCTYPE','EPILOG','INCOMMENT','INPI','VALUE1','VALUE2','CDATA') {
-	print "  statenames[$_] = NULL;\n";
+	print "  ${tagprefix}statenames[$_] = NULL;\n";
       }
       for my $tag (@tags) {
 	my $c = $ctag{$tag};
-	print "  statenames[ROOT_$c] = NULL;\n" if $roottags{$tag};
-	print "  statenames[AL_$c] = NULL;\n";
+	print "  ${tagprefix}statenames[ROOT_${tagprefix}$c] = NULL;\n" if $roottags{$tag};
+	print "  ${tagprefix}statenames[AL_${tagprefix}$c] = NULL;\n";
 	for (split ',',$states{$tag}) {
-	  print "  statenames[$_] = \"$tag\";\n";
+	  print "  ${tagprefix}statenames[$_] = \"$tag\";\n";
 	}
       }
       print "  }\n";
     }
     elsif ( /^FLEXML_DOCTYPES$/ ) {
 
-      my $sysid = ( $uri ? "(\"'$uri'\"|\"\\\"$uri\\\"\")"
-			    : "(\"'$dtd'\"|\"\\\"$dtd\\\"\")" );
+      $sysid = $sysid ?          "(\"'$sysid'\"|\"\\\"$sysid\\\"\")"
+	              : ( $uri ? "(\"'$uri'\"|\"\\\"$uri\\\"\")"
+			       : "(\"'$dtd'\"|\"\\\"$dtd\\\"\")" );
 
       for (keys %roottags) {
 	my $c = $ctag{$_};
 	print " \"<!DOCTYPE\"{S}\"$_\"{S}"
-	    . "SYSTEM{S}" . $sysid . "{s}\">\" SET(ROOT_$c);\n";
+	    . "SYSTEM{S}" . $sysid . "{s}\">\" SET(ROOT_${tagprefix}$c);\n";
 	if ($pubid) {
 	  print " \"<!DOCTYPE\"{S}\"$_\"{S}"
 	      . "PUBLIC{S}(\"'$pubid'\"|\"\\\"$pubid\\\"\"){S}"
-	      . $sysid . "{s}\">\" SET(ROOT_$c);\n";
+	      . $sysid . "{s}\">\" SET(ROOT_${tagprefix}$c);\n";
 	}
       }	  
 
@@ -1245,16 +1368,11 @@ if ($opt{S} or $opt{A}) {
 	  }
 	  $exitswitch .= "  }\n";
 	}
-	
-
-
-
-
 
 	# Start or empty tag: initialise attribute list.
 	print "\n";
 	if ($roottags{$tag}) {
-	  print "<ROOT_$myctag" . ($instates{$tag} ? ",$instates{$tag}" : "");
+	  print "<ROOT_${tagprefix}$myctag" . ($instates{$tag} ? ",$instates{$tag}" : "");
 	}
 	else {
 	  print "<$instates{$tag}";
@@ -1265,9 +1383,9 @@ if ($opt{S} or $opt{A}) {
 	print ">\"<$tag\"{s} {\n";
 
 	for my $attribute (@myattributes) {
-	  print "  A_${myctag}_$catt{$attribute} = " . $initof{"$tag/$attribute"} . ";\n";
+	  print "  AX_${tagprefix}${myctag}_$catt{$attribute} = " . $initof{"$tag/$attribute"} . ";\n";
 	}
-	print "  ENTER(AL_$myctag);\n";
+	print "  ENTER(AL_${tagprefix}$myctag); pushbuffer(0);\n";
 	print "  }\n";
 
 #	print " .       FAIL(\"Unexpected character `%c': `<$tag' expected.\",yytext[0]);\n";
@@ -1277,7 +1395,7 @@ if ($opt{S} or $opt{A}) {
 
 	# Attribute list (of start or empty tag):
 	print "\n";
-	print "<AL_$myctag>{\n";
+	print "<AL_${tagprefix}$myctag>{\n";
 
 	for my $attribute (@myattributes) {
 	  my $type; # set by conditions
@@ -1289,7 +1407,7 @@ if ($opt{S} or $opt{A}) {
 		print " \"$attribute\"{Eq}\"'" . $attdef{"$tag/$attribute"} . "'\""
 		    . " |\n"
 		    . " \"$attribute\"{Eq}\"\\\"" . $attdef{"$tag/$attribute"} . "\\\"\""
-		    . " A_${myctag}_$catt{$attribute}"
+		    . " A_${tagprefix}${myctag}_$catt{$attribute}"
 		    . " = " . $initof{"$tag/$attribute"} . ";\n";
 	    }
 	    else {
@@ -1298,8 +1416,8 @@ if ($opt{S} or $opt{A}) {
 		print " \"$attribute\"{Eq}\"'$alternative'\""
 		    . " |\n"
 		    . " \"$attribute\"{Eq}\"\\\"$alternative\\\"\""
-		    . " A_${myctag}_$catt{$attribute}"
-		    . " = A_${myctag}_$catt{$attribute}_" . variablify($alternative) . ";\n";
+		    . " A_${tagprefix}${myctag}_$catt{$attribute}"
+		    . " = A_${tagprefix}${myctag}_$catt{$attribute}_" . variablify($alternative) . ";\n";
 	      }
 	    }
 	  }
@@ -1310,23 +1428,23 @@ if ($opt{S} or $opt{A}) {
 	    print " \"$attribute\"{Eq}\"'" . $attdef{"$tag/$attribute"} . "'\""
 	        . " |\n"
 		. " \"$attribute\"{Eq}\"\\\"" . $attdef{"$tag/$attribute"} . "\\\"\""
-		. " A_${myctag}_$catt{$attribute}"
+		. " AX_${tagprefix}${myctag}_$catt{$attribute}"
 		. " = " . $initof{"$tag/$attribute"} . ";\n";
 
 	  }
 	  elsif ($type = $literaltype{"$tag/$attribute"}) {
 
 	    # - (non-fixed) literal-type attribute: scan literal string directly, or
-	    print " \"$attribute\"{Eq}\'$type\' BUFFERLITERAL('\\\'',A_${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\'$type\' BUFFERLITERAL('\\\'',AX_${tagprefix}${myctag}_$catt{$attribute});\n";
 
-	    print " \"$attribute\"{Eq}\\\"$type\\\" BUFFERLITERAL('\"',A_${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\\\"$type\\\" BUFFERLITERAL('\"',AX_${tagprefix}${myctag}_$catt{$attribute});\n";
 
 	  }
 	  else {
 
 	    # - (non-fixed non-literal) attribute: scan string with entity expansion.
-	    print " \"$attribute\"{Eq}\\' ENTER(VALUE1); BUFFERSET(A_${myctag}_$catt{$attribute});\n";
-	    print " \"$attribute\"{Eq}\\\" ENTER(VALUE2); BUFFERSET(A_${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\\' ENTER(VALUE1); BUFFERSET(AX_${tagprefix}${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\\\" ENTER(VALUE2); BUFFERSET(AX_${tagprefix}${myctag}_$catt{$attribute});\n";
 
 	  }
 	  print "\n";
@@ -1337,13 +1455,13 @@ if ($opt{S} or $opt{A}) {
 	print " \">\" {\n";
 	for my $attribute (@myattributes) {
 	  if ($required{"$tag/$attribute"}) {
-	    print "  if (!A_$ctag{$tag}_" . variablify($attribute) . ")"
+	    print "  if (!AX_${tagprefix}$ctag{$tag}_" . variablify($attribute) . ")"
 	        . " FAIL(\"Required attribute `$attribute' not set for `$tag' element.\");\n";
 	  }
 	}
-	print "  LEAVE; STag_$myctag();"
-	    . (%inmixed ? ' pushbuffer(pcdata);' : '')
-	    . ($mixed{$tag} ? 'BUFFERSET(pcdata)' : 'pcdata = NULL'). ";"
+	print "  LEAVE; STag_${tagprefix}$myctag();"
+	    . (%inmixed ? ' pushbuffer('."${tagprefix}".'pcdata_ix);' : '')
+	    . ($mixed{$tag} ? 'pushbuffer('."${tagprefix}".'pcdata_ix); BUFFERSET('."${tagprefix}".'pcdata_ix);' : "${tagprefix}".'pcdata_ix = 0'). ";"
 	    . " ENTER($startstate{$tag});\n";
 	print " }\n";
 	#
@@ -1352,16 +1470,16 @@ if ($opt{S} or $opt{A}) {
 	  print " \"/>\" {\n";
 	  for my $attribute (@myattributes) {
 	    if ($required{"$tag/$attribute"}) {
-	      print "  if (!A_$ctag{$tag}_" . variablify($attribute) . ")"
+	      print "  if (!AX_${tagprefix}$ctag{$tag}_" . variablify($attribute) . ")"
 		  . " FAIL(\"Required attribute `$attribute' not set for `$tag' element.\");\n";
 	    }
 	  }
-	  print "  LEAVE; STag_$myctag();"
-	      . (%inmixed ? ' pushbuffer(pcdata);' : '')
-	      . ' pcdata = ' . ($mixed{$tag} ? '""' : 'NULL') . ';'
-	      . " ETag_$myctag();"
-	      . (%inmixed ? ' pcdata = popbuffer();' : '')
-	      . "\n";
+	  print "  LEAVE; STag_${tagprefix}$myctag();"
+	      . (%inmixed ? ' pushbuffer('."${tagprefix}".'pcdata_ix);' : '')
+	      . " ${tagprefix}".'pcdata_ix = 0;'
+	      . " ETag_${tagprefix}$myctag();"
+	      . (%inmixed ? " ${tagprefix}".'pcdata_ix = popbuffer();' : '')
+	      . " popbuffer(); /* attribute */\n";
 	  #
 	  print $exitswitch;
 	  print " }\n";
@@ -1384,8 +1502,10 @@ if ($opt{S} or $opt{A}) {
 	print " \"</$tag\"{s}\">\" {\n";
 	print "  LEAVE;\n";
 	print "  BUFFERDONE;\n" if $mixed{$tag};
-	print "  ETag_$myctag();\n";
-	print "  pcdata = popbuffer();\n" if %inmixed;
+	print "  ETag_${tagprefix}$myctag();\n";
+	print "  ${tagprefix}pcdata_ix = popbuffer();\n" if $mixed{$tag};
+	print "  ${tagprefix}pcdata_ix = popbuffer();\n" if %inmixed;
+	print "  popbuffer(); /* attribute */\n";
 	print $exitswitch;
 	print " }\n";
 
@@ -1400,7 +1520,7 @@ if ($opt{S} or $opt{A}) {
 	# Errors when expecting root tag.
 	if ($roottags{$tag} and $nofail) {
 	  print "\n";
-	  print "<ROOT_$myctag>{\n";
+	  print "<ROOT_${tagprefix}$myctag>{\n";
 	  print " .       FAIL(\"Unexpected character `%c': `$tag' element expected.\",yytext[0]);\n";
 	  print " <<EOF>> FAIL(\"EOF in prolog.\");\n";
 	  print "}\n";
@@ -1422,8 +1542,8 @@ if ($opt{S} or $opt{A}) {
 
       my $sep = $`;
       for (@tags) {
-	print $sep . ($roottags{$_} ? "ROOT_$ctag{$_}," : "")
-            . "AL_$ctag{$_}";
+	print $sep . ($roottags{$_} ? "ROOT_${tagprefix}$ctag{$_}," : "")
+            . "AL_${tagprefix}$ctag{$_}";
 	print ",$states{$_}" if $properempty{$_} or $children{$_};
 	$sep = ',';
       }
@@ -1433,8 +1553,8 @@ if ($opt{S} or $opt{A}) {
     elsif ( /FLEXML_COMMENTS([,>])/ ) {
 
       print "$`"
-	  . join(',', map(($roottags{$_} ? "ROOT_$ctag{$_}," : "")
-			  . "AL_$ctag{$_},$states{$_}", @tags))
+	  . join(',', map(($roottags{$_} ? "ROOT_${tagprefix}$ctag{$_}," : "")
+			  . "AL_${tagprefix}$ctag{$_},$states{$_}", @tags))
 	  . "$1$'";
 
     }
@@ -1478,13 +1598,19 @@ if ($opt{S} or $opt{A}) {
     }
     else {
       s/"\$Id/"\$" "Id/;
+      s/statenames/${tagprefix}statenames/g;
+      s/element_context/${tagprefix}element_context/g;
+      s/parse_err_msg/${tagprefix}parse_err_msg/g;
+      s/rcs_flexml_skeleton/rcs_${tagprefix}flexml_skeleton/g;
+      s/bufferliteral/${tagprefix}bufferliteral/g;
+      s/bufferstack/${tagprefix}bufferstack/g;
       print;
     }
 
   }
 
   close SKELETON || die "$0: Cannot close $SKELETON: $!\n";
-  unless ($opt{A}) {
+  unless ($standalone) {
     close SCANNER || die "$0: Cannot close $SCANNER: $!\n";
   }
 }
@@ -1494,10 +1620,10 @@ if ($opt{S} or $opt{A}) {
 # WRITE APPLICATION.
 
 
-if ($opt{D}) {
+if ($dummy) {
 
   print STDOUT "Writing XML"
-             . ($opt{a} ? "" : " dummy")
+             . ($actions ? "" : " dummy")
              . " application onto `$APPLICATION'.\n" if $verbose;
 
   open APPLICATION, "+>$APPLICATION" || die "$0: Cannot write $APPLICATION: $!\n";
@@ -1519,12 +1645,12 @@ if ($opt{D}) {
 
 }
 
-if ($opt{D} or $opt{A}) {
+if ($dummy or $standalone) {
 
   # Get requested actions.
   if ($ACTIONS) {
 
-    open ACTIONS, "./flexml-act $ACTIONS|" || die "$0: Cannot exec ./flexml-act $ACTIONS: $!\n";
+    open ACTIONS, "$actbin $ACTIONS|" || die "$0: Cannot exec $actbin $ACTIONS: $!\n";
 
     my ($tag,$attribute);
     my @myattributes;
@@ -1537,19 +1663,19 @@ if ($opt{D} or $opt{A}) {
       if ( m/^\#line ([0-9]+)/ ) {
 	$lineno = $1;
       }
-      elsif ( m/^void\s+STag_($Name)\(void\)$/xo ) {
+      elsif ( m/^void\s+STag_${tagprefix}($Name)\(void\)$/xo ) {
 	$tag = $1;
 	die "\"$ACTIONS\", line $lineno: Unknown element `$tag'.\n" unless $ctag{$tag};
 	$startok{$tag} = 'true';
 	@myattributes = (exists $withattr{$tag} ? split /,/,"$withattr{$tag}" : ());
 	$isstart = 'true';
       }
-      elsif ( m|^\}\s+\/\*\s+STag_($Name)\s+\*\/$|xo ) {
+      elsif ( m|^\}\s+\/\*\s+STag_${tagprefix}($Name)\s+\*\/$|xo ) {
 	$tag = undef;
 	@myattributes = ();
 	$isstart = 'true';
       }
-      elsif ( m/^void\s+ETag_($Name)\(void\)$/xo ) {
+      elsif ( m/^void\s+ETag_${tagprefix}($Name)\(void\)$/xo ) {
 	$tag = $1;
 	$endok{$1} = 'true';
 	$isstart = undef;
@@ -1561,24 +1687,18 @@ if ($opt{D} or $opt{A}) {
       # Replace special annotations with C equivalents.
       if ($tag) {
 
-	while ( s/\{($Name)\}/A_$ctag{$tag}_$catt{$1}/x ) {
-	  die "\"$ACTIONS\", line $lineno: Attributes only allowed in start tag.\n"
-	    if not $isstart;
+	while ( s/\{($Name)\}/A_${tagprefix}$ctag{$tag}_$catt{$1}/x ) {
 	  die "\"$ACTIONS\", line $lineno: Unknown attribute `$1' for <$tag>.\n"
 	    if not $atttype{"$tag/$1"};
 	}
 
-	while ( s/\{[!]($Name)\}/AU_$ctag{$tag}_$catt{$1}/x ) {
-	  die "\"$ACTIONS\", line $lineno: Attributes only allowed in start tag.\n"
-	    if not $isstart;
+	while ( s/\{[!]($Name)\}/AU_${tagprefix}$ctag{$tag}_$catt{$1}/x ) {
 	  die "\"$ACTIONS\", line $lineno: Unknown attribute `$1' for <$tag>.\n"
 	    if not $atttype{"$tag/$1"};
 	}
 
 	while ( s|\{($Name)=($Name)\}|
-		  "A_$ctag{$tag}_$catt{$1}_" . variablify($2); |xe ) {
-	  die "\"$ACTIONS\", line $lineno: Attributes only allowed in start tag.\n"
-	    if not $isstart;
+		  "A_${tagprefix}$ctag{$tag}_$catt{$1}_" . variablify($2); |xe ) {
 	  my ($att,$elt) = ($1,$2);
 	  die "\"$ACTIONS\", line $lineno: Unknown attribute $1 for <$tag>.\n"
 	    if not $atttype{"$tag/$1"};
@@ -1586,7 +1706,7 @@ if ($opt{D} or $opt{A}) {
 	    if not $enumtype{"$tag/$att"} =~ m/\b$elt\b/ ;
 	}
 
-	while ( s|\{\#(PCDATA)\}|pcdata| ) {
+	while ( s|\{\#(PCDATA)\}|${tagprefix}pcdata| ) {
 	  die "\"$ACTIONS\", line $lineno: {#PCDATA} only allowed in end tag.\n"
 	    if $isstart;
 	  die "\"$ACTIONS\", line $lineno: {#PCDATA} only permitted in end tag with Mixed contents.\n"
@@ -1596,10 +1716,11 @@ if ($opt{D} or $opt{A}) {
 	die "\"$ACTIONS\", line $lineno: Malformed annotation `$&' in <$tag> action.\n"
 	  if m|\{[^;\s]+\}|o;
       }
+      print STDERR "Action: $_" if $debug;
       print $_;
 
     }
-    close ACTIONS || die "$0: Cannot close pipe to flexml-act: $!\n";
+    close ACTIONS || die "$0: Cannot close pipe to $actbin: $!\n";
     print "\n";
   }
 
@@ -1608,14 +1729,13 @@ if ($opt{D} or $opt{A}) {
 
 }
 
-if ($opt{D}) {
+if ($dummy) {
   close APPLICATION || die "$0: Cannot close $APPLICATION: $!\n";
 }
-elsif ($opt{A}) {
+elsif ($standalone) {
   close SCANNER || die "$0: Cannot close $SCANNER: $!\n";
 }
 
-
 
 
 =pod
@@ -1630,6 +1750,7 @@ B<flexml>
 [B<-ASHDvdnLXV>]
 [B<-s>I<skel>]
 [B<-p>I<pubid>]
+[B<-i>I<init_header>]
 [B<-u>I<uri>]
 [B<-r>I<rootags>]
 [B<-a>I<actions>]
@@ -1674,92 +1795,114 @@ I<Flexml> takes the following options.
 
 =over 4
 
-=item B<-A>
+=item B<--stand-alone>, B<-A> 
 
 Generate a I<stand-alone> scanner application.  If combined with
 B<-a>I<actions> then the application will be named as I<actions> with
 the extension replaced by F<.l>, otherwise it will be in I<name>F<.l>.
 Conflicts with B<-S>, B<-H>, and B<-D>.
 
-=item B<-a> I<actions>
+=item B<--actions> I<actions>, B<-a> I<actions>
 
 Uses the I<actions> file to produce an XML application in the file
 with the same name as I<actions> after replacing the extension with
 F<.c>.  If combined with B<-A> then instead the stand-alone
 application will include the action functions.
 
-=item B<-D>
+=item B<--dummy> B<[>I<app_name>B<]>, B<-D> B<[>I<app_name>B<]>
 
-Generate a dummy application I<name>F<-dummy.c> with just empty
-functions to be called by the XML processor.  If combined with
-B<-a>I<actions> then the application will insert the specified
-actions and be named as I<actions> with the extension replaced by
-F<.c>.  Conflicts with B<-A>; implied by B<-a> unless either of
-B<-SHD> is specified.
+Generate a dummy application with just empty functions to be called by the
+XML processor. If I<app_name> is not specified on the command line, it
+defaults to I<name>F<-dummy.c>.  If combined with B<-a> I<actions> then the
+application will insert the specified actions and be named as I<actions>
+with the extension replaced by F<.c>.  Conflicts with B<-A>; implied by
+B<-a> unless either of B<-SHD> is specified.
 
-=item B<-d>
+=item B<--debug>, B<-d>
 
 Turns on debug mode in the flex scanner and also prints out the
 details of the DTD analysis performed by I<flexml>.
 
-=item B<-H>
+=item B<--header> B<[>I<header_name>B<]>, B<-H> B<[>I<header_name>B<]>
 
-Generate the header file I<name>F<.h>.  Conflicts with B<-A>; on by
+Generate the header file. If the I<header_name> is not specified on the
+command line, defaults to I<name>F<.h>.  Conflicts with B<-A>; on by
 default if none of B<-SHD> specified.
 
-=item B<-L>
+=item B<--lineno>, B<-L>
 
 Makes the XML processor (as produced by I<flex>(1)) count the lines in
 the input and keep it available to XML application actions in the
 integer C<yylineno>.  (This is off by default as the performance
 overhead is significant.)
 
-=item B<-q>
+=item B<--quiet>, B<-q>
 
 Prevents the XML processor (as produced by I<flex>(1)) from reporting 
 the error it runs into on stderr. Instead, users will have to pool for
 error messages with the parse_err_msg() function.
 By default, error messages are written on stderr.
 
-=item B<-n>
+=item B<--dry-run>, B<-n>
 
 "Dry-run": do not produce any of the output files.
 
-=item B<-p> I<pubid>
+=item B<--pubid> I<pubid>, B<-p> I<pubid>
 
 Sets the document type to be C<PUBLIC> with the identifier I<pubid>
 instead of C<SYSTEM>, the default.
 
-=item B<-r> I<roottags>
+=item B<--init_header> I<init_header>, B<-i> I<init_header>
+
+Puts a line containing C<#include "init_header"> in the C<%{...%}> section
+at the top of the generated .l file.  This may be useful for making various
+flex C<#define>s, for example C<YY_INPUT> or C<YY_DECL>.
+
+=item B<--sysid>=I<sysid>
+
+Overrides the C<SYSTEM> id of the accepted DTD. Sometimes usefull when your
+dtd is placed in a subdirectory.
+
+=item B<--root-tags> I<roottags>, B<-r> I<roottags>
 
 Restricts the XML processor to validate only documents with one of the
 root elements listed in the comma-separated I<roottags>.
 
-=item B<-S>
+=item B<--scanner> B<[>I<scanner_name>B<]>, B<-S> B<[>I<scanner_name>B<]>
 
-Generate the scanner I<name>F<.l>.  Conflicts with B<-A>; on by
-default if none of B<-SHD> specified.
+Generate the scanner. If I<scanner_name> is not given on command line, it
+defaults to I<name>F<.l>.  Conflicts with B<-A>; on by default if none of
+B<-SHD> specified.
 
-=item B<-s> I<skel>
+=item B<--skel> I<skel>, B<-s> I<skel>
 
 Use the skeleton scanner I<skel> instead of the default.
 
-=item B<-b> I<stack_size>
+=item B<--act-bin> I<flexml-act>, B<-T> I<flexml-act>
 
-Sets the FLEXML_BUFFERSTACKSIZE to stack_size (100000 by default). Use
-this option when you get an error like "Assertion `next<limit'
-failed".
+This is an internal option mainly used to test versions of flexml not
+installed yet.
 
-=item B<-u> I<uri>
+=item B<--stack-increment> I<stack_increment>, B<-b> I<stack_increment>
+
+Sets the FLEXML_BUFFERSTACKSIZE to stack_increment (100000 by default). This
+controls how much the data stack grows in each realloc().
+
+=item B<--tag-prefix> I<STRING>, B<-O> I<STRING>
+
+Use STRING to differentiate multiple versions of flexml in the same C
+code, just like the -P flex argument.
+
+=item B<--uri> I<uri>, B<-u> I<uri>
 
 Sets the URI of the DTD, used in the C<DOCTYPE> header, to the
 specified I<uri> (the default is the DTD name).
 
-=item B<-v>
+=item B<--verbose>, B<-v>
 
 Be verbose: echo each DTD declaration (after parameter expansion).
 
-=item B<-V>
+=item B<--version>, B<-V>
 
 Print the version of I<flexml> and exit.
 
@@ -1910,11 +2053,6 @@ in particular C<xml:spaces> should be supported.
 
 =item *
 
-The XML processor currently uses a fixed-size buffer to read
-C<pcdata>.  It should not.
-
-=item *
-
 The DTD parser is presently a perl hack so it may parse some DTDs
 badly; in particular the expansion of parameter entities may not
 conform fully to the XML specification.
@@ -1948,7 +2086,7 @@ REC-xml-1998-0210).
 
 =head1 AUTHOR
 
-I<Flexml> was written by Kristoffer Hgsbro Rose,
+I<Flexml> was written by Kristoffer Rose,
 E<lt>C<krisrose@debian.org>E<gt>.
 
 =head1 COPYRIGHT
